from __future__ import annotations

from typing import Any, Optional

from vial.errors import ServerError
from vial.types import Request


class RequestContext:
    _INSTANCE: Optional[RequestContext] = None

    def __init__(self, request: Request) -> None:
        self.request = request

    def __enter__(self) -> RequestContext:
        RequestContext._INSTANCE = self
        return self

    def __exit__(self, *_: Any) -> None:
        RequestContext._INSTANCE = None

    @classmethod
    def active(cls) -> RequestContext:
        if not cls._INSTANCE:
            raise ServerError("Not currently within a request")
        return cls._INSTANCE


def get() -> Request:
    return RequestContext.active().request



import logging
from logging import Formatter, Logger, StreamHandler


class LoggerFactory:

    DEFAULT_FORMAT = "[%(levelname)s] %(asctime)s %(filename)s.%(funcName)s: %(message)30s"

    @classmethod
    def get(cls, name: str) -> Logger:
        log = logging.getLogger(name)
        log.setLevel(logging.DEBUG)
        log.propagate = False
        log.addHandler(cls.get_handler())
        return log

    @classmethod
    def get_handler(cls) -> StreamHandler:
        handler = StreamHandler()
        handler.setFormatter(Formatter(cls.DEFAULT_FORMAT))
        return handler



from vial.middleware import MiddlewareAPI
from vial.parsers import ParserAPI
from vial.routes import RoutingAPI


class Resource(RoutingAPI, ParserAPI, MiddlewareAPI):
    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name



from dataclasses import dataclass, field
from enum import Enum, auto
from http import HTTPStatus
from typing import Any, Dict, Iterator, List, Mapping, MutableMapping, Optional, Protocol, TypeVar, Union

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class MultiDict(MutableMapping[K, List[V]]):  # pylint: disable=too-many-ancestors
    def __init__(self, values: Optional[MutableMapping[K, List[V]]] = None) -> None:
        super().__init__()
        self._values = values or {}

    def __delitem__(self, key: K) -> None:
        del self._values[key]

    def __getitem__(self, key: K) -> List[V]:
        return self._values[key]

    def get_first(self, key: K) -> V:
        return self._values[key][0]

    def extend(self, key: K, value: List[V]) -> None:
        existing_values = self._values.get(key)
        if existing_values is not None:
            existing_values.extend(value)
        else:
            self._values[key] = value

    def add(self, key: K, value: V) -> None:
        existing_values = self._values.get(key)
        if existing_values is not None:
            existing_values.append(value)
        else:
            self._values[key] = [value]

    def __iter__(self) -> Iterator[K]:
        return iter(self._values)

    def __len__(self) -> int:
        return len(self._values)

    def __setitem__(self, key: K, value: List[V]) -> None:
        self._values[key] = value


class Json(Protocol):
    @staticmethod
    def dumps(value: Any) -> str:
        pass

    @staticmethod
    def loads(value: str) -> Any:
        pass


@dataclass
class CognitoIdentity:
    cognito_identity_id: str
    cognito_identity_pool_id: str


@dataclass
class MobileClient:
    installation_id: str
    app_title: str
    app_version_name: str
    app_version_code: str
    app_package_name: str


@dataclass
class ClientContext:
    client: MobileClient
    custom: Dict[str, Any]
    env: Dict[str, Any]


@dataclass
class LambdaContext:
    function_name: str
    function_version: str
    invoked_function_arn: str
    memory_limit_in_mb: int
    aws_request_id: str
    log_group_name: str
    log_stream_name: str
    identity: Optional[CognitoIdentity] = None
    client_context: Optional[ClientContext] = None

    @staticmethod
    def get_remaining_time_in_millis() -> int:
        return 0


class HTTPMethod(Enum):
    GET = auto()
    PUT = auto()
    PATCH = auto()
    POST = auto()
    DELETE = auto()
    OPTIONS = auto()
    TRACE = auto()


@dataclass
class LambdaEvent:
    event: Dict[str, Any]
    context: LambdaContext


@dataclass
class Request(LambdaEvent):
    method: HTTPMethod
    resource: str
    path: str
    headers: MultiDict[str, str]
    query_parameters: MultiDict[str, str]
    body: Optional[str]


@dataclass
class Response:
    body: Optional[Union[Mapping[str, Any], List[Any], str]] = None
    headers: Dict[str, str] = field(default_factory=dict)
    status: Union[HTTPStatus, int] = HTTPStatus.OK



from http import HTTPStatus


class ServerError(Exception):
    status = HTTPStatus.INTERNAL_SERVER_ERROR


class BadRequestError(ServerError):
    status = HTTPStatus.BAD_REQUEST


class UnauthorizedError(ServerError):
    status = HTTPStatus.UNAUTHORIZED


class ForbiddenError(ServerError):
    status = HTTPStatus.FORBIDDEN


class NotFoundError(ServerError):
    status = HTTPStatus.NOT_FOUND


class MethodNotAllowedError(ServerError):
    status = HTTPStatus.METHOD_NOT_ALLOWED



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Mapping, Tuple

from vial.parsers import KeywordParser, Parser
from vial.types import HTTPMethod, T


@dataclass(frozen=True)
class Route:
    resource: str
    path: str
    method: HTTPMethod
    variables: Mapping[str, Parser]
    function: Callable[..., Any]
    metadata: Mapping[str, Any]


class RoutingAPI:
    name: str
    param_parser: KeywordParser

    def __init__(self) -> None:
        super().__init__()
        self.routes: Dict[str, Dict[HTTPMethod, Route]] = defaultdict(dict)

    def post(self, path: str, **kwargs: Any) -> Callable[[Callable[..., T]], Callable[..., T]]:
        return self.route(path, [HTTPMethod.POST], **kwargs)

    def put(self, path: str, **kwargs: Any) -> Callable[[Callable[..., T]], Callable[..., T]]:
        return self.route(path, [HTTPMethod.PUT], **kwargs)

    def patch(self, path: str, **kwargs: Any) -> Callable[[Callable[..., T]], Callable[..., T]]:
        return self.route(path, [HTTPMethod.PATCH], **kwargs)

    def delete(self, path: str, **kwargs: Any) -> Callable[[Callable[..., T]], Callable[..., T]]:
        return self.route(path, [HTTPMethod.DELETE], **kwargs)

    def get(self, path: str, **kwargs: Any) -> Callable[[Callable[..., T]], Callable[..., T]]:
        return self.route(path, [HTTPMethod.GET], **kwargs)

    def route(
        self, path: str, methods: Iterable[HTTPMethod] = tuple(HTTPMethod), **kwargs: Any
    ) -> Callable[[Callable[..., T]], Callable[..., T]]:
        def registrar(function: Callable[..., T]) -> Callable[..., T]:
            for method in methods:
                self._register_route(path, method, function, kwargs)
            return function

        return registrar

    def register_routes(self, other: RoutingAPI) -> None:
        for resource, routes in other.routes.items():
            self.routes[resource].update(routes)

    def _register_route(
        self, path: str, method: HTTPMethod, function: Callable[..., T], metadata: Mapping[str, Any]
    ) -> None:
        route = self._build_route(path, method, function, metadata)
        self.routes[route.path][method] = route

    def _build_route(
        self, path: str, method: HTTPMethod, function: Callable[..., Any], metadata: Mapping[str, Any]
    ) -> Route:
        parsed_components: List[str] = []
        variables: Dict[str, Parser] = {}
        self._parse_components(path.split("/"), parsed_components, variables)
        return Route(self.name, "/".join(parsed_components), method, variables, function, metadata)

    def _parse_components(
        self, components: List[str], parsed_components: List[str], variables: Dict[str, Parser]
    ) -> None:
        for component in components:
            if component.startswith("{") and component.endswith("}"):
                name, parser = self._build_variable(component)
                variables[name] = parser
                parsed_components.append(f"{{{name}}}")
            else:
                parsed_components.append(component)

    def _build_variable(self, component: str) -> Tuple[str, Parser]:
        placeholder = component[1:-1]
        name_and_parser = placeholder.split(":")
        if len(name_and_parser) == 1:
            return name_and_parser[0], str
        return name_and_parser[0], self.param_parser.get(name_and_parser[1])



import json
from http import HTTPStatus
from typing import Any, Dict, Mapping, Type

from vial.errors import MethodNotAllowedError, NotFoundError, ServerError
from vial.loggers import LoggerFactory
from vial.middleware import CallChain, MiddlewareAPI, MiddlewareChain
from vial.parsers import ParserAPI
from vial.request import RequestContext
from vial.resources import Resource
from vial.routes import Route, RoutingAPI
from vial.types import HTTPMethod, Json, LambdaContext, MultiDict, Request, Response


class NativeJson(Json):
    @staticmethod
    def dumps(value: Any) -> str:
        return json.dumps(value)

    @staticmethod
    def loads(value: str) -> Any:
        return json.loads(value)


class RouteResolver:
    def __call__(self, resources: Mapping[str, Mapping[HTTPMethod, Route]], request: Request) -> Route:
        defined_routes = resources.get(request.resource)
        if not defined_routes:
            raise NotFoundError(request.resource)

        route = defined_routes.get(request.method)
        if not route:
            raise MethodNotAllowedError(request.method.name)

        return route


class RouteInvoker:
    def __call__(self, route: Route, request: Request) -> Response:
        args = self._build_args(route, request)
        result = route.function(*args.values())
        return self._to_response(result)

    @staticmethod
    def _to_response(result: Any) -> Response:
        if isinstance(result, Response):
            return result
        if result is None or not isinstance(result, tuple):
            return Response(result)
        return Response(*result)

    @staticmethod
    def _build_args(route: Route, request: Request) -> Mapping[str, Any]:
        if not route.variables:
            return {}
        path_params: Mapping[str, str] = request.event["pathParameters"]
        args: Dict[str, Any] = {}
        for name, parser in route.variables.items():
            args[name] = parser(path_params[name])
        return args


class ErrorHandler:

    DEFAULT_STATUSES = {
        Exception: HTTPStatus.INTERNAL_SERVER_ERROR,
        ValueError: HTTPStatus.BAD_REQUEST,
        NotImplementedError: HTTPStatus.NOT_IMPLEMENTED,
    }

    def __call__(self, error: Exception) -> Response:
        return Response({"message": str(error)}, status=self._get_status_code(error))

    def _get_status_code(self, error: Exception) -> HTTPStatus:
        if isinstance(error, ServerError):
            return error.status
        return self._get_native_status_code(error)

    def _get_native_status_code(self, error: Exception) -> HTTPStatus:
        for error_type in type(error).__mro__:
            if issubclass(error_type, Exception):
                status = self.DEFAULT_STATUSES.get(error_type)
                if status:
                    return status
        return HTTPStatus.INTERNAL_SERVER_ERROR


class Vial(RoutingAPI, ParserAPI, MiddlewareAPI):

    route_resolver_class = RouteResolver

    invoker_class = RouteInvoker

    logger_factory_class = LoggerFactory

    error_handler_class = ErrorHandler

    json_class: Type[Json] = NativeJson

    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name
        self.route_resolver = self.route_resolver_class()
        self.error_handler = self.error_handler_class()
        self.invoker = self.invoker_class()
        self.json = self.json_class()
        self.logger = self.logger_factory_class.get(name)

    def register_resource(self, app: Resource) -> None:
        ParserAPI.register_parsers(self, app)
        RoutingAPI.register_routes(self, app)
        MiddlewareAPI.register_middlewares(self, app)

    def __call__(self, event: Dict[str, Any], context: LambdaContext) -> Mapping[str, Any]:
        request = self._build_request(event, context)
        with RequestContext(request):
            response = self._handle_request(request)
            return self._to_lambda_response(response)

    def _handle_request(self, request: Request) -> Response:
        try:
            route = self.route_resolver(self.routes, request)
            return self._build_invocation_chain(route)(request)
        except Exception as e:  # pylint: disable=broad-except
            self.logger.exception("Encountered uncaught exception")
            return self.error_handler(e)

    def _build_invocation_chain(self, route: Route) -> CallChain:
        all_middleware = self.registered_middleware[self.name] + self.registered_middleware[route.resource]

        def route_invocation(event: Request) -> Response:
            return self.invoker(route, event)

        handler = MiddlewareChain(all_middleware[-1], route_invocation)
        for middleware in reversed(all_middleware[0:-1]):
            handler = MiddlewareChain(middleware, handler)
        return handler

    def _to_lambda_response(self, response: Response) -> Mapping[str, Any]:
        if not isinstance(response.body, str):
            body = self.json.dumps(response.body) if response.body is not None else None
        else:
            body = response.body
        return {"headers": response.headers, "statusCode": response.status, "body": body}

    @staticmethod
    def _build_request(event: Dict[str, Any], context: LambdaContext) -> Request:
        return Request(
            event,
            context,
            HTTPMethod[event["httpMethod"]],
            event["resource"],
            event["path"],
            MultiDict(event["multiValueHeaders"]),
            MultiDict(event["multiValueQueryStringParameters"]),
            event["body"],
        )



from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Type, Union
from urllib.parse import parse_qs, urlparse

from vial.app import NativeJson, Vial
from vial.errors import NotFoundError
from vial.types import HTTPMethod, Json, LambdaContext, Response


@dataclass
class Match:
    route: str
    path_params: Mapping[str, str]
    query_params: Mapping[str, List[str]]


class RouteMatcher:
    def __init__(self, routes: List[str]) -> None:
        self.routes = sorted(routes)

    def match(self, url: str) -> Match:
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.query, keep_blank_values=True)
        path = self._remove_trailing_slash(parsed_url.path)
        return self._find_match(url, path.split("/"), query_params)

    def _find_match(self, url: str, parts: List[str], query_params: Mapping[str, List[str]]) -> Match:
        for route_url in self.routes:
            url_parts = route_url.split("/")
            if len(parts) != len(url_parts):
                continue

            path_params = self._match_path(parts, url_parts)
            if path_params is not None:
                return Match(route_url, path_params, query_params)
        raise NotFoundError(f"No matching route found for {url}")

    @staticmethod
    def _match_path(parts: List[str], url_parts: List[str]) -> Optional[Mapping[str, str]]:
        path_params: Dict[str, str] = {}
        for left, right in zip(parts, url_parts):
            if right.startswith("{") and right.endswith("}"):
                path_params[right[1:-1]] = left
                continue

            if left != right:
                return None
        return path_params

    @staticmethod
    def _remove_trailing_slash(path: str) -> str:
        if path != "/" and path.endswith("/"):
            return path[:-1]
        return path


class Gateway:

    json_class: Type[Json] = NativeJson

    def __init__(self, app: Vial) -> None:
        self.app = app
        self.json = self.json_class()
        self.matcher = RouteMatcher(list(app.routes))

    def get(self, path: str, headers: Optional[Mapping[str, str]] = None) -> Response:
        request = self.build_request(HTTPMethod.GET, path, None, headers)
        response = self.app(request, self.get_context())
        body: Optional[str] = response["body"]
        return Response(self.json.loads(body) if body else None, response["headers"], response["statusCode"])

    def build_request(
        self,
        method: HTTPMethod,
        path: str,
        body: Optional[str],
        headers: Optional[Mapping[str, Union[str, Sequence[str]]]] = None,
    ) -> Dict[str, Any]:
        match = self.matcher.match(path)
        return {
            "httpMethod": method.name,
            "resource": match.route,
            "path": path,
            "multiValueHeaders": self._build_headers(headers or {}),
            "multiValueQueryStringParameters": match.query_params,
            "pathParameters": match.path_params,
            "body": body,
        }

    @staticmethod
    def _build_headers(headers: Mapping[str, Union[str, Sequence[str]]]) -> Mapping[str, List[str]]:
        multi_headers: Mapping[str, List[str]] = defaultdict(list)
        for name, value in headers.items():
            if isinstance(value, Iterable):
                multi_headers[name].extend(value)
            else:
                multi_headers[name].append(value)
        return multi_headers

    @staticmethod
    def get_context() -> LambdaContext:
        return LambdaContext("vial-test", "1", "arn:vial-test", 128, "1", "vial-test-log", "vial-test-log")



from __future__ import annotations

from collections import defaultdict
from typing import Callable, Dict, List, Protocol

from vial.types import Request, Response, T


class CallChain(Protocol):
    def __call__(self, event: Request) -> Response:
        pass


class MiddlewareChain:
    def __init__(self, handler: Callable[[Request, CallChain], Response], next_call: CallChain) -> None:
        self.handler = handler
        self.next_call = next_call

    def __call__(self, event: Request) -> Response:
        return self.handler(event, self.next_call)


class MiddlewareAPI:
    name: str

    def __init__(self) -> None:
        super().__init__()
        self.registered_middleware: Dict[str, List[Callable[[Request, CallChain], T]]] = defaultdict(list)

    def middleware(self, function: Callable[[Request, CallChain], T]) -> Callable[[Request, CallChain], T]:
        self.register_middleware(function)
        return function

    def register_middleware(self, middleware: Callable[[Request, CallChain], T]) -> None:
        self.registered_middleware[self.name].append(middleware)

    def register_middlewares(self, other: MiddlewareAPI) -> None:
        self.registered_middleware.update(other.registered_middleware)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vial']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pyvial',
    'version': '0.7.0',
    'description': 'A micro web framework for AWS Lambda.',
    'long_description': '# Vial\nA micro web framework for AWS Lambda.\n\n## Installation\nTo add vial to your project, run the following command:\n```\npip install pyvial\n```\n\n## Usage\n### Entry Point\nThe main entry point of the application is always the `Vial#__call__` function. When deploying to AWS Lambda,\nthe Lambda handler should point to the `Vial` object in whichever file it\'s defined in. As an example:\n```\nfrom vial.app import Vial\n\napp = Vial(__name__)\n```\nIf this code snippet is defined in an `app.py` file, the handler would be `app.app`.\n\n### Basic API\n```\nfrom typing import Mapping\nfrom vial.app import Vial\n\napp = Vial(__name__)\n\n\n@app.get("/hello-world")\ndef hello_world() -> Mapping[str, str]:\n    return {"hello": "world"}\n```\n\n### Current Request\nThe current request is tracked within a contextual object that wraps the lambda request. It can be accessed like so:\n```\nfrom typing import Mapping\n\nfrom vial import request\nfrom vial.app import Vial\n\napp = Vial(__name__)\n\n\n@app.get("/hello-world")\ndef hello_world() -> Mapping[str, List[str]]:\n    query_params = request.get().query_parameters\n    if not query_params:\n        raise ValueError("Must provide at least one query parameter")\n    return dict(query_params)\n```\n\n### Path Parameters\nYou can define path parameters like this:\n```\n@app.get("/users/{user_id}")\ndef get_user(user_id: str) -> User:\n    return user_service.get(user_id)\n```\n\nVial supports some path parameter parsing as part of the invocation process. For example when using a UUID\nas a path parameter, Vial can convert it from a string to a UUID automatically:\n```\nfrom uuid import UUID\n\n@app.get("/users/{user_id:uuid}")\ndef get_user(user_id: UUID) -> User:\n    return user_service.get(user_id)\n```\n\nThe following parsers are supported by default:\n\n| Parser        | Type              |\n| ------------- | ----------------- |\n| `str`         | `str`             |\n| `bool`        | `bool`            |\n| `int`         | `int`             |\n| `float`       | `float`           |\n| `decimal`     | `decimal.Decimal` |\n| `uuid`        | `uuid.UUID`       |\n\nYou can register your own parser like this:\n```\n@app.parser("list")\ndef list_parser(value: str) -> List[str]:\n    return [value]\n\n\n@app.get("/users/{user_id:list}")\ndef get_user(user_ids: List[str]) -> List[User]:\n    return [user_service.get(user_id) for user_id in user_ids]\n```\nAs parsers are bound directly to the registered route function, they have to be defined before the route\nfunction that uses one is registered.\n\n## Resources\nAs your application grows, you may want to split certain functionality amongst resources, similar to\nblueprints of other popular frameworks like Flask.\n\nYou can define a resource like this:\n```\n# store.py\nfrom vial.resources import Resource\n\napp = Resource(__name__)\n\n\n@app.get("/stores/{store_id}")\ndef get_store(store_id: str) -> Store:\n    return store_service.get(store_id)\n\n\n# app.py\nfrom stores import app as stores_app\n\n\napp = Vial(__name__)\n\napp.register_resource(stores_app)\n```\n\n## Middleware\nYou can register middleware functions to be executed before / after route invocations. All middleware is scoped to\nwhere it\'s registered. A middleware function registered with the `Vial` instance is scoped to all routes within\nthe application, but a function registered with a `Resource` instance will only be invoked for routes defined in\nthat specific resource.\n\nBelow is an example of registering a middleware to log route invocation:\n```\nimport logging\nfrom typing import Mapping\nfrom vial.app import Vial\n\napp = Vial(__name__)\n\nlogger = logging.getLogger("my-app")\n\n@app.middleware\ndef log_events(event: Request, chain: CallChain) -> Response:\n    logger.info("Began execution of %s", event.context)\n    try:\n        return chain(event)\n    finally:\n        logger.info("Completed execution of %s", event.context)\n\n\n@app.get("/hello-world")\ndef hello_world() -> Mapping[str, str]:\n    return {"hello": "world"}\n```\n\n\n## Json Encoding\nYou can customize how Vial serializes / deserializes JSON objects by passing a custom encoder. The below\nexample shows how to substitute the native JSON module with another library like `simplejson`:\n```\nimport simplejson\nfrom vial.app import Vial, Json\n\n\nclass SimpleJson(Json):\n    @staticmethod\n    def dumps(value: Any) -> str:\n        return simplejson.dumps(value)\n\n    @staticmethod\n    def loads(value: str) -> Any:\n        return simplejson.loads(value)\n\nclass JsonVial:\n    json_class = SimpleJson\n\n\napp = SimpleJsonVial()\n```\n',
    'author': 'Michael Dimchuk',
    'author_email': 'michaeldimchuk@gmail.com',
    'maintainer': 'Michael Dimchuk',
    'maintainer_email': 'michaeldimchuk@gmail.com',
    'url': 'https://github.com/michaeldimchuk/pyvial',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from decimal import Decimal
from typing import Any, Callable, Mapping
from uuid import UUID

from vial.types import T

Parser = Callable[[str], Any]


class KeywordParser:
    """
    A path parameter parser that converts string parameters into more specific types
    prior to being injected into route handler functions.
    """

    DEFAULT_PARSERS: Mapping[str, Parser] = {
        "str": str,
        "bool": bool,
        "int": int,
        "float": float,
        "decimal": Decimal,
        "uuid": UUID,
    }

    def __init__(self) -> None:
        self.parsers = dict(self.DEFAULT_PARSERS)

    def get(self, name: str) -> Parser:
        parser = self.parsers.get(name)
        if not parser:
            raise ValueError(f"Parser '{name}' is not registered.")
        return parser

    def register(self, name: str, parser: Parser) -> None:
        if name in self.parsers:
            raise ValueError(f"Parser for keyword {name} already exists")
        self.parsers[name] = parser


class ParserAPI:

    parser_class = KeywordParser

    def __init__(self) -> None:
        super().__init__()
        self.param_parser = self.parser_class()

    def parser(self, name: str) -> Callable[[Callable[[str], T]], Callable[[str], T]]:
        def registration_function(function: Callable[[str], T]) -> Callable[[str], T]:
            self.register_parser(name, function)
            return function

        return registration_function

    def register_parser(self, name: str, parser: Callable[[str], T]) -> None:
        self.param_parser.register(name, parser)

    def register_parsers(self, other: ParserAPI) -> None:
        self.param_parser.parsers.update(other.param_parser.parsers)
